/*
 * $Id: Bld.java,v 1.3 2002/06/08 12:23:38 kenta Exp $
 *
 * Copyright (c) 2002 Kenta Cho. All rights reserved.
 */

import com.nttdocomo.ui.*;
import com.nttdocomo.opt.ui.j3d2.*;
import java.util.Random;

/**
 * BuildingRacer main class.
 *
 * @author cs8k-cyu@asahi-net.or.jp
 * @version $Revision: 1.3 $
 */
public class Bld extends IApplication {
  private NullCanvas canvas;
  private C3DGraphics g;
  private int scWidth, scHeight, scCenterX, scCenterY;

  private final int fx1 = 65536;

  private final int colb = 49152;

  private boolean backLight = true;
  /**
   * Handle backlight.
   */
  private void setBackLight() {
    int blStat;
    if ( backLight ) {
      blStat = PhoneSystem.ATTR_BACKLIGHT_ON;
    } else {
      blStat = PhoneSystem.ATTR_BACKLIGHT_OFF;
    }
    PhoneSystem.setAttribute(PhoneSystem.DEV_BACKLIGHT, blStat);
  }

  /**
   * Initialize.
   */
  private void init() {
    canvas = new NullCanvas();
    canvas.setSoftLabel(Frame.SOFT_KEY_2, "Exit");
    scWidth = canvas.getWidth();
    scHeight = canvas.getHeight();
    scCenterX = scWidth/2;
    scCenterY = scHeight/2;
    setBackLight();
    Display.setCurrent(canvas);
    initTitle();
  }

  /**
   * Make images for textures.
   */
  private void makeTexture() {
    byte[] bp = new byte[16*16*4];
    for ( int y=0 ; y<16 ; y++ ) {
      for ( int x=0 ; x<16 ; x++ ) {
        int i = (x+y*16)*4;
        if ( x == 0 || y == 0 ) {
          bp[i] = bp[i+1] = bp[i+2] = bp[i+3] = (byte)255;
        } else {
          bp[i+3] = (byte)128;
          bp[i] = bp[i+1] = bp[i+2] = (byte)(256-(x+y)*8);
        }
      }
    }
    g.bindTexture(g.C3DL_TEXTURE_2D, 1);
    g.texImage2D(g.C3DL_TEXTURE_2D, 0,
                 g.C3DL_RGBA8, 16, 16, 0, g.C3DL_RGBA, g.C3DL_UNSIGNED_BYTE,
                 bp);

    for ( int i=0 ; i<6 ; i++ ) {
      makeStringTexture(2+i*2, bb[i][0], false);
      makeStringTexture(3+i*2, bb[i][1], true);
    }
    for ( int i=0 ; i<10 ; i++ ) {
      makeStringTexture(20+i, String.valueOf(i), false);
    }
    makeStringTexture(30, "score", false);
  }

  /**
   * Init canvas after C3DGraphics instance becomes available.
   */
  private void initCanvas() {
    g.set3DMode(true);

    g.viewport(0, 0, scWidth, scHeight);
    g.clearColor(0, 0, 0, fx1);

    g.enable(g.C3DL_TEXTURE_2D);
    makeTexture();

    g.blendFunc(g.C3DL_SRC_ALPHA, g.C3DL_ONE);
    g.enable(g.C3DL_BLEND);
    g.enable(g.C3DL_LINE_SMOOTH);
    g.disable(g.C3DL_DEPTH_TEST);
    g.disable(g.C3DL_CULL_FACE);

    g.matrixMode(g.C3DL_PROJECTION);
    g.loadIdentity();
    int aspect = fx1 * scWidth / scHeight;
    g.perspective(45*fx1, aspect, fx1/256, 512*fx1);

    g.matrixMode(g.C3DL_MODELVIEW);
    g.loadIdentity();

    g.enable(g.C3DL_LIGHTING);
    g.light (g.C3DL_LIGHT0, g.C3DL_AMBIENT, new int[]{ fx1/6, fx1/6, fx1/6, fx1 });
    g.light (g.C3DL_LIGHT0, g.C3DL_POSITION, new int[]{ 0, 0, 0, fx1 });
    g.light (g.C3DL_LIGHT0, g.C3DL_SPOT_CUTOFF, new int[]{ fx1*180 });
    g.enable(g.C3DL_LIGHT0);
    g.material(g.C3DL_FRONT_AND_BACK, g.C3DL_AMBIENT_AND_DIFFUSE, new int[]{ fx1, fx1, fx1, fx1 });
    g.enable(g.C3DL_COLOR_MATERIAL);
    g.colorMaterial(g.C3DL_FRONT_AND_BACK, g.C3DL_AMBIENT_AND_DIFFUSE);
  }

  private final int LN = 1;

  private final short[] lt = {
    0x7bed,0x6bee,0x7927,0x6b6e,0x79a7,0x79a4,0x796f,0x5bed,0x2492,0x126f,0x5bad,0x4927,0x5fed,0x5ffd,0x7b6f,0x7be4,0x7b7f,0x6bad,0x79cf,0x7492,0x5b6f,0x5b6a,0x5ffa,0x5aad,0x5bd2,0x72a7,0x7b6f,0x2492,0x73e7,0x73cf,0x5bc9,0x79cf,0x79ef,0x7249,0x7bef,0x7bcf,0x3493,0x6496,
  };

  /**
   * Make a texuture with text.
   */
  private void makeStringTexture(int tn, String s, boolean r) {
    int w, l = s.length();
    if ( l == 1 ) {
      w = 4;
    } else {
      w = 32;
    }
    byte[] bp = new byte[w*8*4];
    char c;
    short p, m;
    int k;
    byte rc, gc, bc;
    for ( int i=0, x=0 ; i<l ; i++, x+=4 ) {
      c = s.charAt(i);
      rc = gc = bc = (byte)255;
      if ( c == ' ' ) continue;
      else if ( c == '[' ) {
        c = (char)('9'-'0'+1 + 'z'-'a'+1);
        gc = bc = (byte)128;
      } else if ( c == ']' ) {
        c = (char)('9'-'0'+1 + 'z'-'a'+1 + 1);
        gc = bc = (byte)128;
      }
      else if ( c >= 'a' && c <='z' ) {
        c -= 'a';
      } else {
        c = (char)(c-'0' + 'z'-'a'+1);
        bc = (byte)128;
      }
      p = lt[c];
      m = 0x4000;
      for ( int j=0 ; j<15 ; j++ ) {
        k = ((j%3)+x + (j/3)*w) * 4;
        if ( r ) k += (8-l)*4*4;
        if ( (p&m) != 0 ) {
          bp[k] = rc; bp[k+1] = gc; bp[k+2] = bc;
          bp[k+3] = (byte)255;
        } else {
          bp[k] = bp[k+1] = bp[k+2] = bp[k+3] = (byte)0;
        }
        m >>= 1;
      }
    }
    g.bindTexture(g.C3DL_TEXTURE_2D, tn);
    g.texImage2D(g.C3DL_TEXTURE_2D, 0,
                 g.C3DL_RGBA8, w, 8, 0, g.C3DL_RGBA, g.C3DL_UNSIGNED_BYTE,
                 bp);
  }

  // Bulidings data.
  private final int BM = 9, BW = 3;
  private final int BSW = 65536;

  private int[] bx = new int[BM];
  private int[] by = new int[BM];
  private int[] bxw = new int[BM];
  private int[] byw = new int[BM];
  private int[] brx = new int[BM];
  private int[] bry = new int[BM];
  private int[] brxw = new int[BM];
  private int[] bryw = new int[BM];
  private int[] bh = new int[BM];
  private int[] bs = new int[BM];

  // Fragments data.
  private final int FM = 8;

  private int[] fx = new int[FM];
  private int[] fy = new int[FM];
  private int[] fz = new int[FM];
  private int[] fpx = new int[FM];
  private int[] fpy = new int[FM];
  private int[] fpz = new int[FM];
  private int[] fmx = new int[FM];
  private int[] fmy = new int[FM];
  private int[] fmz = new int[FM];
  private int[] fc = new int[FM];

  // Car data.
  private int cx, cy, cd, cmd, cs, ch, cbx, cby, co;

  // Viewpoint data.
  private int ex, ey, ez, eax, eay, eaz;
  private int sx, sy, sz, sax, say, saz;

  // Game status (0-title  1-in game 2-crashed 3-gamvover).
  private int gs;
  private int gc;
  private int sc;

  /**
   * Initialize car, buildings, fragments.
   */
  private void initGameState() {
    cx = BSW*BW/2; cy = BSW*BW/2; cd = 0; cmd = 0; cs = 0; co = 0;
    ch = 13312;
    cbx = cx>>16; cby = cy>>16;

    for ( int y=0 ; y<BW ; y++ ) {
      for ( int x=0 ; x<BW ; x++ ) {
        addBld(x, y);
      }
    }
    bs[4] = 0;
    for ( int i=0 ; i<FM ; i++ ) {
      fc[i] = -1;
    }
  }

  /**
   * Start title.
   */
  private void initTitle() {
    gs = 0; gc = 0;
    initGameState();
  }

  /**
   * Start game.
   */
  private void initInGame() {
    gs = 1; gc = 0;
    sc = -1;
    initGameState();
  }

  /**
   * Start gameover.
   */
  private void initGameover() {
    gs = 3; gc = 0;
    initGameState();
  }

  /**
   * Add building.
   */
  private void addBld(int x, int y) {
    int ax, ay;
    if ( x < 0 ) ax = BW + (x%BW) - 1;
    else ax = x%BW;
    if ( y < 0 ) ay = BW + (y%BW) - 1;
    else ay = y%BW;
    int i = ax + ay*BW;
    bx[i] = BSW*x + randN(BSW);
    by[i] = BSW*y + randN(BSW);
    bxw[i] = randN(BSW/6) + BSW/10;
    byw[i] = randN(BSW/6) + BSW/10;
    bh[i] = randN(BSW/6) + BSW/10;
    bs[i] = 32;
  }

  /**
   * Player's car crashd, gameover.
   */
  private void carCrashed() {
    gs = 2; gc = 0;
    for ( int i=0 ; i<8 ; i++ ) {
      addFrag(cx, 128, cy, randNS(2048), randN(1024), randNS(2048));
    }
    eax = cx+(sin(cd)>>1)+randNS(2048);
    eay = (ch<<1)+randNS(2048);
    eaz = cy+(cos(cd)>>1)+randNS(2048);
  }

  /**
   * Grow building size.
   */
  private void moveBlds() {
    int x, y, xw, yw, s;
    for ( int i=0 ; i<BM ; i++ ) {
      if ( bs[i] < 32 ) continue;
      if ( gs == 1 ) {
        if ( bs[i] < 3840 ) {
          bs[i] += (4096-bs[i])>>3;
        } else {
          bs[i] += 72+(sc<<1);
        }
        s = bs[i];
        brxw[i] = xw = (bxw[i]*s)>>12; bryw[i] = yw = (byw[i]*s)>>12;
        brx[i] = x = bx[i]-(xw>>1); bry[i] = y = by[i]-(yw>>1);
        if ( cx > x && cx < x+xw && cy > y && cy < y+yw ) {
          carCrashed();
        }
      } else {
        bs[i] -= bs[i]>>2;
      }
    }
  }

  // Handle fragments.
  private int fi = FM;

  private void addFrag(int x, int y, int z, int mx, int my, int mz) {
    fi--; if ( fi < 0 ) fi = FM-1;
    fx[fi] = fpx[fi] = x; fy[fi] = fpy[fi] = y; fz[fi] = fpz[fi] = z;
    fmx[fi] = mx; fmy[fi] = my; fmz[fi] = mz;
    fc[fi] = randN(16)+16;
  }

  private void moveFrags() {
    for ( int i=0 ; i<FM ; i++ ) {
      if ( fc[i] < 0 ) continue;
      fpx[i] = fx[i]; fpy[i] = fy[i]; fpz[i] = fz[i];
      fx[i] += fmx[i]; fy[i] += fmy[i]; fz[i] += fmz[i];
      fmx[i] -= fmx[i]>>5; fmz[i] -= fmz[i]>>5;
      fc[i]--;
    }
  }

  private final int SMAX = 1600, MDMAX = fx1*6;

  /**
   * Move my car.
   */
  private void moveCar() {
    int bx, by;
    cx += (sin(cd)*cs)>>15;
    cy += (cos(cd)*cs)>>15;
    cd += cmd;
    cmd -= cmd>>3;
    cs -= cs>>8;

    bx = cx>>16; by = cy>>16;
    if ( bx != cbx ) {
      if ( bx < cbx ) {
        for ( int y=by-(BW/2) ; y<=by+(BW/2) ; y++ ) {
          addBld(bx-1, y);
        }
      } else {
        for ( int y=by-(BW/2) ; y<=by+(BW/2) ; y++ ) {
          addBld(bx+1, y);
        }
      }
      cbx = bx;
    }
    if ( by != cby ) {
      if ( by < cby ) {
        for ( int x=bx-(BW/2) ; x<=bx+(BW/2) ; x++ ) {
          addBld(x, by-1);
        }
      } else {
        for ( int x=bx-(BW/2) ; x<=bx+(BW/2) ; x++ ) {
          addBld(x, by+1);
        }
      }
      cby = by;
    }

    if ( (co==1 || co==3) && randN(7)==0 ) {
      addFrag(cx, 128, cy,
              ((-sin(cd)*(cs+(SMAX>>2)))>>15)+randNS(1024),
              randN(1024),
              ((-cos(cd)*(cs+(SMAX>>2)))>>15)+randNS(1024));
    }

    eax = cx-(sin(cd)>>1); eay = ch; eaz = cy-(cos(cd)>>1);
    sax = cx; say = 1280; saz = cy;

    if ( (gc%20) == 0 ) {
      sc++;
    }
  }

  /**
   * Move an eye position and a looking position.
   */
  private void moveViewpoint() {
    ex += (eax-ex)>>2; ey += (eay-ey)>>2; ez += (eaz-ez)>>2;
    sx += (sax-sx)>>1; sy += (say-sy)>>1; sz += (saz-sz)>>1;
  }

  // Title and gameover messages.
  private final String bb[][] = {
    {"building", "racer"},
    {"push[7]", "to start"},
    {"[arrow]", "steering"},
    {"[7]", "accel"},
    {"[4]", "brake"},
    {"  game", "over  "}
  };

  private int dx, dy, dd, dt;

  /**
   * Set the billboard for title and gameover.
   */
  private void moveBillboard() {
    if ( (gc%92) == 0 ) {
      int mc = (gc/92)%5;
      if ( gs == 3 ) {
        if ( gc > 0 ) return;
        mc = 5;
      }
      dt = 2+mc*2;
      dx = randNS(fx1*8); dy = randNS(fx1*8);
      cx = dx; cy = dy;

      saz = dy;
      eay = fx1/16;
      if ( dd == 0 ) {
        dd = 1;
        eax = sax = dx+23500;
        eaz = dy + fx1;
      } else {
        dd = 0;
        eax = sax = dx+9000;
        eaz = dy - fx1;
      }
    } else {
      if ( dd == 0 ) {
        eaz += fx1/372;
      } else {
        eaz -= fx1/372;
      }
    }
  }

  /**
   * Move 1 frame.
   */
  private void move() {
    switch ( gs ) {
    case 0:
      moveBillboard();
      break;
    case 1:
      moveCar();
      moveBlds();
      moveFrags();
      break;
    case 2:
      moveBlds();
      moveFrags();
      if ( gc > 40 ) {
        initGameover();
        return;
      }
      break;
    case 3:
      moveBillboard();
      if ( gc > 160 ) {
        initTitle();
        return;
      }
      break;
    }
    moveViewpoint();
    gc++;
  }

  private boolean pPrsd = true, sPrsd = true;

  /**
   * Check a key pad state.
   */
  private void handleKeyState() {
    int key = canvas.getKeypadState();
    if ( (key&(1<<Display.KEY_SOFT2)) != 0 ) {
      terminate();
    }
    if ( (key&(1<<Display.KEY_POUND)) == 0 ) {
      pPrsd = false;
    } else if ( !pPrsd ) {
      pPrsd = true;
      if ( backLight ) {
        backLight = false;
      } else {
        backLight = true;
      }
      setBackLight();
    }
    switch ( gs ) {
    case 0:
      if ( (key&(1<<Display.KEY_7)) == 0 ) {
        sPrsd = false;
      } else if ( !sPrsd ) {
        sPrsd = true;
        initInGame();
      }
      break;
    case 1:
      if ( (key&(1<<Display.KEY_7)) != 0 ) {
        cs += (SMAX-cs)>>4;
        co = 3;
      } else if ( (key&(1<<Display.KEY_4)) != 0 ) {
        cs -= cs>>3;
        co = 1;
      } else {
        co = 2;
      }
      if ( (key&(1<<Display.KEY_UP)) != 0 ) {
        cmd += (MDMAX-cmd)>>co;
      }
      if ( (key&(1<<Display.KEY_DOWN)) != 0 ) {
        cmd += (-MDMAX-cmd)>>co;
      }
      if ( (key&(1<<Display.KEY_LEFT)) != 0 ) {
        if ( ch > 8192 ) ch -= 512;
      }
      if ( (key&(1<<Display.KEY_RIGHT)) != 0 ) {
        if ( ch < 32768 ) ch += 512;
      }
      break;
    case 2:
      if ( (key&(1<<Display.KEY_7)) == 0 ) {
        sPrsd = false;
      } else if ( !sPrsd ) {
        sPrsd = true;
        initGameover();
      }
      break;
    case 3:
      if ( (key&(1<<Display.KEY_7)) == 0 ) {
        sPrsd = false;
      } else if ( !sPrsd ) {
        sPrsd = true;
        initTitle();
      }
      break;
    }
  }

  /**
   * Draw bulidings.
   */
  private void drawBlds() {
    g.bindTexture(g.C3DL_TEXTURE_2D, 1);

    int x, y, xw, yw, h, s;
    for ( int i=0 ; i<BM ; i++ ) {
      if ( bs[i] < 32 ) continue;
      s = bs[i];
      x = brx[i]; y = bry[i];
      xw = brxw[i]; yw = bryw[i];
      h = (bh[i]*s)>>12;

      g.begin(g.C3DL_QUADS);

      g.color(colb, colb/2, colb/2, fx1);
      g.normal(0, 0, -fx1);
      g.texCoord(0, 0);
      g.vertex(x, 0, y);
      g.texCoord(0, h<<4);
      g.vertex(x, h, y);
      g.texCoord(xw<<4, h<<4);
      g.vertex(x+xw, h, y);
      g.texCoord(xw<<4, 0);
      g.vertex(x+xw, 0, y);

      g.color(colb/2, colb, colb/2, fx1);
      g.normal(0, 0, fx1);
      g.texCoord(xw<<4, 0);
      g.vertex(x, 0, y+yw);
      g.texCoord(0, 0);
      g.vertex(x+xw, 0, y+yw);
      g.texCoord(0, h<<4);
      g.vertex(x+xw, h, y+yw);
      g.texCoord(xw<<4, h<<4);
      g.vertex(x, h, y+yw);

      g.color(colb/2, colb/2, colb, fx1);
      g.normal(-fx1, 0, 0);
      g.texCoord(0, 0);
      g.vertex(x, 0, y);
      g.texCoord(yw<<4, 0);
      g.vertex(x, 0, y+yw);
      g.texCoord(yw<<4, h<<4);
      g.vertex(x, h, y+yw);
      g.texCoord(0, h<<4);
      g.vertex(x, h, y);

      g.color(colb, colb, colb/2, fx1);
      g.normal(fx1, 0, 0);
      g.texCoord(yw<<4, 0);
      g.vertex(x+xw, 0, y);
      g.texCoord(yw<<4, h<<4);
      g.vertex(x+xw, h, y);
      g.texCoord(0, h<<4);
      g.vertex(x+xw, h, y+yw);
      g.texCoord(0, 0);
      g.vertex(x+xw, 0, y+yw);

      g.end();
    }
  }

  /**
   * Draw fragments.
   */
  private void drawFrags() {
    g.color(fx1, fx1, 0, fx1/3*2);
    g.begin(g.C3DL_LINES);
    for ( int i=0 ; i<FM ; i++ ) {
      if ( fc[i] < 0 ) continue;
      g.vertex(fx[i], fy[i], fz[i]);
      g.vertex(fpx[i], fpy[i], fpz[i]);
    }
    g.end();
  }

  /**
   * Draw my car.
   */
  private void drawCar() {
    g.bindTexture(g.C3DL_TEXTURE_2D, 1);
    g.color(fx1, fx1, fx1, fx1);

    g.begin(g.C3DL_TRIANGLE_FAN);

    g.normal(0, fx1, 0);
    g.texCoord(0, 0);
    g.vertex(cx+(sin(cd)>>4), 256, cy+(cos(cd)>>4));
    g.normal(0, fx1, 0);
    g.texCoord(0, fx1>>1);
    g.vertex(cx+(sin(cd+fx1*120)>>5), 128, cy+(cos(cd+fx1*120)>>5));
    g.normal(0, fx1, 0);
    g.texCoord(fx1>>1, fx1>>1);
    g.vertex(cx+(sin(cd+fx1*180)>>6), 1024, cy+(cos(cd+fx1*180)>>6));
    g.normal(0, fx1, 0);
    g.texCoord(fx1>>1, 0);
    g.vertex(cx+(sin(cd+fx1*240)>>5), 128, cy+(cos(cd+fx1*240)>>5));

    g.end();
  }

  /**
   * Draw a ground texture.
   */
  private void drawGround() {
    g.bindTexture(g.C3DL_TEXTURE_2D, 1);
    g.color(colb, colb, colb, fx1);

    g.begin(g.C3DL_QUADS);

    int gx = cx, gy = cy;
    int tx = (cx&65535)<<2, ty = (cy&65535)<<2;

    g.normal(0, fx1, 0);
    g.texCoord(tx, ty);
    g.vertex(gx, 0, gy);
    g.texCoord(fx1*16+tx, ty);
    g.vertex(gx+fx1*4, 0, gy);
    g.texCoord(fx1*16+tx, fx1*16+ty);
    g.vertex(gx+fx1*4, 0, gy+fx1*4);
    g.texCoord(tx, fx1*16+ty);
    g.vertex(gx, 0, gy+fx1*4);

    g.normal(0, fx1, 0);
    g.texCoord(tx, ty);
    g.vertex(gx-fx1*4, 0, gy);
    g.texCoord(fx1*16+tx, ty);
    g.vertex(gx, 0, gy);
    g.texCoord(fx1*16+tx, fx1*16+ty);
    g.vertex(gx, 0, gy+fx1*4);
    g.texCoord(tx, fx1*16+ty);
    g.vertex(gx-fx1*4, 0, gy+fx1*4);

    g.normal(0, fx1, 0);
    g.texCoord(tx, ty);
    g.vertex(gx, 0, gy-fx1*4);
    g.texCoord(fx1*16+tx, ty);
    g.vertex(gx+fx1*4, 0, gy-fx1*4);
    g.texCoord(fx1*16+tx, fx1*16+ty);
    g.vertex(gx+fx1*4, 0, gy);
    g.texCoord(tx, fx1*16+ty);
    g.vertex(gx, 0, gy);

    g.normal(0, fx1, 0);
    g.texCoord(tx, ty);
    g.vertex(gx-fx1*4, 0, gy-fx1*4);
    g.texCoord(fx1*16+tx, ty);
    g.vertex(gx, 0, gy-fx1*4);
    g.texCoord(fx1*16+tx, fx1*16+ty);
    g.vertex(gx, 0, gy);
    g.texCoord(tx, fx1*16+ty);
    g.vertex(gx-fx1*4, 0, gy);

    g.end();
  }

  /**
   * Draw a billboard for title.
   */
  private void drawBillboard() {
    g.color(fx1, fx1, fx1, fx1);

    switch ( dd ) {
    case 0:
      g.bindTexture(g.C3DL_TEXTURE_2D, dt);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(fx1, 0);
      g.vertex(dx, fx1/16*2, dy);
      g.texCoord(0, 0);
      g.vertex(dx+fx1/2, fx1/16*2, dy);
      g.texCoord(0, fx1);
      g.vertex(dx+fx1/2, fx1/16, dy);
      g.texCoord(fx1, fx1);
      g.vertex(dx, fx1/16, dy);
      g.end();
      g.bindTexture(g.C3DL_TEXTURE_2D, dt+1);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(fx1, 0);
      g.vertex(dx-fx1/4, fx1/16, dy);
      g.texCoord(0, 0);
      g.vertex(dx-fx1/4+fx1/2, fx1/16, dy);
      g.texCoord(0, fx1);
      g.vertex(dx-fx1/4+fx1/2, 0, dy);
      g.texCoord(fx1, fx1);
      g.vertex(dx-fx1/4, 0, dy);
      g.end();
      break;
    case 1:
      g.bindTexture(g.C3DL_TEXTURE_2D, dt);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(0, 0);
      g.vertex(dx, fx1/16*2, dy);
      g.texCoord(fx1, 0);
      g.vertex(dx+fx1/2, fx1/16*2, dy);
      g.texCoord(fx1, fx1);
      g.vertex(dx+fx1/2, fx1/16, dy);
      g.texCoord(0, fx1);
      g.vertex(dx, fx1/16, dy);
      g.end();
      g.bindTexture(g.C3DL_TEXTURE_2D, dt+1);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(0, 0);
      g.vertex(dx+fx1/4, fx1/16, dy);
      g.texCoord(fx1, 0);
      g.vertex(dx+fx1/4+fx1/2, fx1/16, dy);
      g.texCoord(fx1, fx1);
      g.vertex(dx+fx1/4+fx1/2, 0, dy);
      g.texCoord(0, fx1);
      g.vertex(dx+fx1/4, 0, dy);
      g.end();
      break;
    }
  }

  /**
   * Draw score.
   */
  private void drawScore() {
    g.color(fx1, fx1, fx1, fx1);

    int x = fx1/18*6, s = sc;

    do {
      g.bindTexture(g.C3DL_TEXTURE_2D, s%10+20);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(0, 0);
      g.vertex(x, fx1/16*6, -fx1);
      g.texCoord(fx1, 0);
      g.vertex(x+fx1/16*6/8, fx1/16*6, -fx1);
      g.texCoord(fx1, fx1);
      g.vertex(x+fx1/16*6/8, fx1/16*4, -fx1);
      g.texCoord(0, fx1);
      g.vertex(x, fx1/16*4, -fx1);
      g.end();
      s /= 10; x -= fx1/16*6/8;
    } while ( s > 0 );

    g.bindTexture(g.C3DL_TEXTURE_2D, 30);
    g.begin(g.C3DL_TRIANGLE_FAN);
    g.texCoord(0, 0);
    g.vertex(-fx1/16*6, fx1/16*6, -fx1);
    g.texCoord(fx1, 0);
    g.vertex(0, fx1/16*6, -fx1);
    g.texCoord(fx1, fx1);
    g.vertex(0, fx1/16*4, -fx1);
    g.texCoord(0, fx1);
    g.vertex(-fx1/16*6, fx1/16*4, -fx1);
    g.end();
  }

  /**
   * Draw to the canvas.
   */
  private void draw() {
    g.set3DMode(true);

    g.clear(g.C3DL_COLOR_BUFFER);

    g.pushMatrix();
    g.rotate(-90*fx1, 0, 0, fx1);

    if ( gs != 0 ) {
      g.enable(g.C3DL_TEXTURE_2D);
      g.disable(g.C3DL_LIGHTING);
      drawScore();
    }

    g.lookAt(ex, ey, ez, sx, sy, sz,
             -sin(cmd<<3), cos(cmd<<3), 0);

    switch (gs) {
    case 0:
    case 3:
      g.enable(g.C3DL_TEXTURE_2D);
      g.disable(g.C3DL_LIGHTING);
      drawBillboard();
      g.enable(g.C3DL_LIGHTING);
      drawGround();
      break;
    case 1:
      g.enable(g.C3DL_LIGHTING);
      drawGround();
      drawCar();
      drawBlds();
      g.disable(g.C3DL_TEXTURE_2D);
      g.disable(g.C3DL_LIGHTING);
      drawFrags();
      break;
    case 2:
      g.enable(g.C3DL_LIGHTING);
      drawGround();
      drawBlds();
      g.disable(g.C3DL_TEXTURE_2D);
      g.disable(g.C3DL_LIGHTING);
      drawFrags();
      break;
    }

    g.popMatrix();
    g.flush();
  }

  /**
   * Main loop. (20 frames/sec)
   */
  public void run() {
    long prvTicks, nowTicks;
    int frame;
    prvTicks = System.currentTimeMillis();
    for ( ; ; ) {
      nowTicks = System.currentTimeMillis();
      frame = (int)(nowTicks-prvTicks)/50;
      if ( frame <= 0 ) {
        frame = 1;
        prvTicks += 50;
        while ( System.currentTimeMillis() < prvTicks ) {}
      } else {
        prvTicks = nowTicks;
        if ( frame > 4 ) {
          frame = 4;
        }
      }
      if ( g == null ) {
        g = canvas.getC3DGraphics();
        if ( g != null ) {
          initCanvas();
        } else {
          continue;
        }
      }
      for ( ; frame>0 ; frame-- ) {
        move();
        handleKeyState();
      }
      g.lock();
      draw();
      g.unlock(true);
    }
  }

  /*
   * Random util methods.
   */
  private static Random rnd = new Random();

  private int randN(int n) {
    return Math.abs(rnd.nextInt()%n);
  }

  private int randNS(int n) {
    return rnd.nextInt()%n;
  }

  /*
   * SinCos util methods.
   */
  private int[] sctbl = new int[1024+256];

  private void initSinCos() {
    for ( int i=0 ; i<1024+256 ; i++ ) {
      sctbl[i] = C3DMath.sin(i*(360*fx1/1024));
    }
  }

  private int sin(int d) {
    return sctbl[((d>>6)/360)&1023];
  }

  private int cos(int d) {
    return sctbl[(((d>>6)/360)&1023) + 256];
  }

  /**
   * Bootstrap.
   */
  public void start() {
    initSinCos();
    init();
    run();
  }
}
